#!/usr/bin/perl

use strict;
use warnings;
use File::Copy;

my $line;			# each line read from the input circuit file
my $n = 0;			# the number of this line
my $error = 0;		# if there is a syntax error in this line
my $input = 0;		# the number of inputs
my $output = 0;		# the number of outputs
my @not_in;			# the input wires of NOT gates
my @not_out;		# the output wires of NOT gates

# open AND-OR-NOT circuit file for read
open(AON_FILE,"<$ARGV[0].txt") || die "$!";		

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

# open dualrail circuit file for write
open(DR_FILE,">$ARGV[0].dualrail.txt") || die "$!";

print DR_FILE "# compiled from AND-OR-NOT circuit to dual-rail AND-OR circuit\n";

# read each line of AND-OR-NOT circuit file
while (($error == 0) && ($line = <AON_FILE>)) {

	$n++;
	$error = 1;

	# if this line is comment, copy it to the dualrail circuit file
	if ($line =~ /^\#/) {
		$error = 0;
		print DR_FILE $line;
	}

	# if this line is empty, copy it to the dualrail circuit file
	if ($line =~ /^(\s+)$/) {
		$error = 0;
		print DR_FILE $line;
	}

	# if this line defines an input or an output
	if ($line =~ /^(INPUT|OUTPUT)\s*\(\s*(\d+)\s*\)(\s*|\s*\#.*)$/) {

		$error = 0;

		my $wire = $1;				# INPUT or OUTPUT
		my $number = $2;			# wire number
		my $name;					# wire name (if not noted, use x + wire number as default)

		# use xi as default name for the ith input
		if ($wire eq "INPUT") {	$input++;	$name = "x".$input;	} 
		# use yi as default name for the ith output
		else {	$output++;	$name = "y".$output;	}

		if ($line =~ /\#\s(\w+)/) {	$name = $1;	}

		# each input/output wire is replaced by two wires representing logic OFF (0) and logic ON (1) seperately
		print DR_FILE $wire,"(",$number*2,")\t # ",$name,"^0\n";
		print DR_FILE $wire,"(",$number*2+1,")\t # ",$name,"^1\n";
	}

	# if this line defines a gate
	if ($line =~ /^(\d+)\s*\=\s*(NOT|AND|OR|NAND|NOR)\s*\(([\s|\d|,]+)\)\s*$/) {

		$error = 0;

		# extract gate information from this line               

		my $out = $1;	# output wire of this gate
		my $gate = $2;	# gate type
 		my @in;			# input wires of this gate

		# one input for NOT gate; multiple inputs for any other gates
		if ($gate eq "NOT") {
		 $in[0] = $3;
		} else {
		 @in = split /,/,$3;
		}
		for (my $i=0; $i<=$#in; $i++) {
			if ($in[$i] =~ /^\s*(\d+)\s*$/) {
				$in[$i] = $1;
			} else {$error = 1;}
		}

		# if it is a NOT gate, remember the wires that should be swapped
		if ($gate eq "NOT") {
			push(@not_out,$out*2,$out*2+1);
			push(@not_in,$in[0]*2+1,$in[0]*2);
		}

		# if it is an AND gate
		if ($gate eq "AND") {

			# output wire being OFF is the OR operation of input wires being OFF
			print DR_FILE $out*2," = OR(",$in[0]*2;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2; 
			}
			print DR_FILE ")\n";

			# output wire being ON is the AND operation of input wires being ON
			print DR_FILE $out*2+1," = AND(",$in[0]*2+1;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2+1; 
			}
			print DR_FILE ")\n";
		}
		
		# if it is an OR gate
		if ($gate eq "OR") {

			# output wire being OFF is the AND operation of input wires being OFF
			print DR_FILE $out*2," = AND(",$in[0]*2;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2; 
			}
			print DR_FILE ")\n";

			# output wire being ON is the OR operation of input wires being ON
			print DR_FILE $out*2+1," = OR(",$in[0]*2+1;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2+1; 
			}
			print DR_FILE ")\n";
		}

		# if it is a NAND gate
		if ($gate eq "NAND") {

			# output wire being OFF is the AND operation of input wires being ON
			print DR_FILE $out*2," = AND(",$in[0]*2+1;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2+1; 
			}
			print DR_FILE ")\n";

			# output wire being ON is the OR operation of input wires being OFF
			print DR_FILE $out*2+1," = OR(",$in[0]*2;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2; 
			}
			print DR_FILE ")\n";
		}

		# if it is a NOR gate
		if ($gate eq "NOR") {

			# output wire being OFF is the OR operation of input wires being ON
			print DR_FILE $out*2," = OR(",$in[0]*2+1;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2+1; 
			}
			print DR_FILE ")\n";

			# output wire being ON is the AND operation of input wires being OFF
			print DR_FILE $out*2+1," = AND(",$in[0]*2;
			for (my $i=1; $i<=$#in; $i++) {
				print DR_FILE ", ",$in[$i]*2; 
			}
			print DR_FILE ")\n";
		}
	}

	if ($error == 1) {
		print "Syntax error in line ",$n,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$n,":\n",$line;
	}
}
close(AON_FILE);
close(DR_FILE);
close(ERROR_FILE);

# search for all wires that should be swapped by NOT gates and swap them

my $swap = 1;

while ($swap == 1) {

	$swap = 0;
	open(DR_FILE_old,"<$ARGV[0].dualrail.txt") || die "$!";
	open(DR_FILE_new,">$ARGV[0].dualrail_new.txt") || die "$!";
	while($line = <DR_FILE_old>){
		for (my $i=0; $i<=$#not_in; $i++) {
			if ($line =~ /$not_out[$i](\,|\))/) {
				$swap = 1;
				$line =~ s/$not_out[$i]/$not_in[$i]/g;
			}
		}
		print DR_FILE_new $line;
	}
	close(DR_FILE_old);
	close(DR_FILE_new);
	unlink("$ARGV[0].dualrail.txt");
	copy("$ARGV[0].dualrail_new.txt","$ARGV[0].dualrail.txt") or die "$!";
	unlink("$ARGV[0].dualrail_new.txt");
}