#!/usr/bin/perl

use strict;
use warnings;
use File::Copy;

my $line;		# each line read from the input circuit file
my $nl = 0;		# the number of this line
my $error = 0;	# if there is a syntax error in this line
my @wire;		# (wire type, upstrem connetions, downstream connections, notes)
my $input = 0;	# number of inputs
my $rep = 0;	# number of reporters

# open an AND-OR circuit file for read
open(AO_FILE,"<$ARGV[0].txt") || die "$!";		

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

# open a seesaw circuit file for write
open(SEESAW_FILE,">$ARGV[0].seesaw.txt") || die "$!";

print SEESAW_FILE "# compiled from AND-OR circuit to seesaw circuit\n";

# read each line of the AND-OR circuit file
while (($error == 0) && ($line = <AO_FILE>)) {

	$nl++;
	$error = 1;

	# if this line is comment, copy it to the seesaw circuit file
	if ($line =~ /^\#/) {
		$error = 0;
		print SEESAW_FILE $line;
	}

	# if this line is empty, skip it
	if ($line =~ /^(\s+)$/) {
		$error = 0;
	}

	# if this line defines an input
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)(\s*|\s*\#.*)$/) {

		$error = 0;

		my $n = $1;												# wire number
		if ($line =~ /\#\s(.+)/) {	${$wire[$n]}[3] = $1;	}	# wire name if noted
	}

	# if this line defines an output
	if ($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)(\s*|\s*\#.*)$/) {

		$error = 0;

		my $n = $1;												# wire number
		if ($line =~ /\#\s(.+)/) {	${$wire[$n]}[3] = $1;	}	# wire name if noted
	}

	# if this line defines a gate
	if ($line =~ /^(\d+)\s*\=\s*(AND|OR)\s*\(([\s|\d|,]+)\)\s*$/) {

		$error = 0;

		my $n = $1;												# wire number
		if ($2 eq "AND") {	${$wire[$n]}[0] = 1;	}			# wire type 1: output of an AND gate
		if ($2 eq "OR") {	${$wire[$n]}[0] = 2;	}			# wire type 2: output of an OR gate
		@{${$wire[$n]}[1]} = split /,/,$3;						# upstream connections
		for (my $i=0; $i<=$#{${$wire[$n]}[1]}; $i++) {
			if (${${$wire[$n]}[1]}[$i] =~ /^\s*(\d+)\s*$/) {
				${${$wire[$n]}[1]}[$i] = $1;
			} else {$error = 1;}
		}

		# save this gate as each upstream gate's downstream connection
		for (my $i=0; $i<=$#{${$wire[$n]}[1]}; $i++) {
			push(@{${$wire[${${$wire[$n]}[1]}[$i]]}[2]},$n);
		}
	}

	if ($error == 1) {
		print "Syntax error in line ",$nl,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$nl,":\n",$line;
	}
}

print SEESAW_FILE "\n";

close(AO_FILE);
close(ERROR_FILE);

# write the input signals to the seesaw circuit file 
for (my $i=0; $i<=$#wire; $i++) {

	# if this wire has no upstream connections but downstream connections, it is an input wire
	if (($#{${$wire[$i]}[1]} < 0) && ($#{${$wire[$i]}[2]} >= 0)) {

		# the corresponding seesaw input signal either go through a fanout gate or not 
		# depends on the number of downstream connections
		if ($#{${$wire[$i]}[2]} == 0) {
			print SEESAW_FILE "INPUT(",$i,") = w[",$i*2+1,",",${${$wire[$i]}[2]}[0]*2,"]"; 
		} else {
			print SEESAW_FILE "INPUT(",$i,") = w[",$i*2,",",$i*2+1,"]";
		}

		# if this input wire has notes, copy the notes to seesaw circuit file
		if ($#{$wire[$i]} == 3) {
			print SEESAW_FILE "\t # ",${$wire[$i]}[3],"\n";
		} else {
			print SEESAW_FILE "\n";
		}
	}
}

print SEESAW_FILE "\n";

# write the output signals to the seesaw circuit file 

for (my $i=0; $i<=$#wire; $i++) {

	# if this wire has upstream connections but no downstream connections, it is an output wire 
	if (($#{${$wire[$i]}[1]} >= 0) && ($#{${$wire[$i]}[2]} < 0)) {

		# add one reporter
		$rep = $rep + 1; 

		# connect this output wire to the reporter
		push(@{${$wire[$i]}[2]},$#wire+$rep);

		# the corresponding seesaw output signal is transtrated to a fluorescence signal from a reporter
		print SEESAW_FILE "OUTPUT(",$i,") = Fluor[",($#wire+$rep)*2,"]";

		# if this output wire has notes, copy the notes to seesaw circuit file
		if ($#{$wire[$i]} == 3) {
			print SEESAW_FILE "\t # ",${$wire[$i]}[3],"\n";
		} else {
			print SEESAW_FILE "\n";
		}
	} 
}

print SEESAW_FILE "\n";

# write the input fanout gates to the seesaw circuit file 
for (my $i=0; $i<=$#wire; $i++) {

	# if this wire has no upstream connections but multiple downstream connections, it needs a fanout gate
	if (($#{${$wire[$i]}[1]} < 0) && ($#{${$wire[$i]}[2]} > 0)) {
		print SEESAW_FILE "inputfanout[",$i*2+1,",",$i*2,",{",${${$wire[$i]}[2]}[0]*2;
		for (my $j=1; $j<=$#{${$wire[$i]}[2]}; $j++) {
			print SEESAW_FILE ",",${${$wire[$i]}[2]}[$j]*2;
		}
		print SEESAW_FILE "}]\n";
	}
}

print SEESAW_FILE "\n";

# write the seesaw AND/OR gates to the seesaw circuit file 
for (my $i=0; $i<=$#wire; $i++) {

	# if this wire has upstream connections, it is the output wire of an AND/OR gate
	if ($#{${$wire[$i]}[1]} >= 0) {

		# write the gate type
		print SEESAW_FILE "seesaw";
		if (${$wire[$i]}[0] == 1) {	print SEESAW_FILE "AND[";	}
		if (${$wire[$i]}[0] == 2) {	print SEESAW_FILE "OR[";	}

		# write the two seesaw gates translated from one AND/OR gate
		print SEESAW_FILE $i*2,",",$i*2+1;

		# write the wires connected to the left side of the first seesaw gate
		print SEESAW_FILE ",{",${${$wire[$i]}[1]}[0]*2+1;
		for (my $j=1; $j<=$#{${$wire[$i]}[1]}; $j++) {
			print SEESAW_FILE ",",${${$wire[$i]}[1]}[$j]*2+1;
		}

		# write the wires connected to the right side of the second seesaw gate
		print SEESAW_FILE "},{",${${$wire[$i]}[2]}[0]*2;
		for (my $j=1; $j<=$#{${$wire[$i]}[2]}; $j++) {
			print SEESAW_FILE ",",${${$wire[$i]}[2]}[$j]*2;
		}
		print SEESAW_FILE "}]\n";
	} 
}

print SEESAW_FILE "\n";

# write the reporter gates to the seesaw circuit file 
for (my $i=1; $i<=$rep; $i++) {
	for (my $j=0; $j<=$#wire; $j++) {
		if (($#{${$wire[$j]}[2]} == 0) && (${${$wire[$j]}[2]}[0] == $#wire+$i)) {
			print SEESAW_FILE "reporter[",($#wire+$i)*2,",",$j*2+1,"]\n";
		}
	} 
}

close(SEESAW_FILE);