#!/usr/bin/perl

use strict;
use warnings;

my $toehold = "TCT";												# toehold sequence
my $toehold_c = "AGA";												# complementary toehold sequence
my $recognition = 15;												# length of recognition domain
my $clamp = "CA";													# clamp sequence
my $clamp_c = "TG";													# complementary clamp sequence
my @fluor = ("ROX","FAM","TYE563","TYE665","Cy5.5");				# fluorophore names
my @quenc = ("RQ","FQ","RQ","RQ","RQ");								# quencher names
my @fluorIDT = ("3ROXN","36-FAM","3TYE563","3TYE665","3Cy55Sp");	# IDT fluorophore names
my @quencIDT = ("5IAbRQ","5IAbFQ","5IAbRQ","5IAbRQ","5IAbRQ");		# IDT quencher names
my $Sf = "CATTTTTTTTTTTCA";											# Sf domain sequence used for fuels

my $line;		# each line read from the input circuit file
my $nl = 0;		# the number of this line
my $error = 0;	# if there is a syntax error in this line
my @seq;		# recognition sequences assigned for each seesaw gate
my @seq_c;		# complementary recognition sequences assigned for each seesaw gate
my $n = 0;		# number of used fluorophore/quencher 
my @gate;		# mark the used gate base strands
my @wire;		# mark the used signal strands

# open a seesaw circuit file for read
open(SEESAW_FILE,"<$ARGV[0].seesaw.txt") || die "$!";

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

# open a recognition domain sequence file for read
open(SEQ_FILE,"<sequences".$recognition."clamp.txt") || die "$!";

# open a seesaw sequence file for write
open(SEESAW_SEQ_FILE,">$ARGV[0].seesaw.sequence.txt") || die "$!";

print SEESAW_SEQ_FILE "# compiled from seesaw circuit to DNA sequences\n";

# read the seesaw circuit file, assign one recognition domain sequence for each seesaw gate
while (($error == 0) && ($line = <SEESAW_FILE>)) {

	$nl++;
	$error = 1;

	# if this line is comment, copy it to the sequence file
	if ($line =~ /^(\#)/) {
		$error = 0;
		print SEESAW_SEQ_FILE $line;
	}

	# if this line is empty, skip it
	if ($line =~ /^(\s+)$/) {
		$error = 0;
	}

	# if this line defines an input, assign a recognition sequence for the left side of the input strand
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/) {	
		$error = 0;
		chomp($seq[$2] = <SEQ_FILE>);	
	}

	# if this line defines an output, skip it
	if (($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*Fluor\s*\[\s*(\d+)\s*\](\s*|\s*\#.*)$/) ||
		($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/)) {	
		$error = 0;
	}

	# if this line defines an input fanout gate, assign a recognition sequence associated the seesaw node
	if ($line =~ /^inputfanout\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {	
		$error = 0;
		chomp($seq[$1] = <SEQ_FILE>);	

		my @fanout = split /,/,$3;
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] !~ /^\s*(\d+)\s*$/) {	$error = 1;	}
		}
	}

	# if this line defines a seesaw AND/OR gate, assign two recognition sequences associated the pair of seesaw nodes
	if ($line =~ /^seesaw(AND|OR)\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/){	
		$error = 0;
		chomp($seq[$2] = <SEQ_FILE>);	
		chomp($seq[$3] = <SEQ_FILE>);

		my @fanin = split /,/,$4;
		my @fanout = split /,/,$5;

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] !~ /^\s*(\d+)\s*$/) {	$error = 1;	}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] !~ /^\s*(\d+)\s*$/) {	$error = 1;	}
		}
	}

	# if this line defines a reporter gate, assign a recognition sequence associated the seesaw node
	if ($line =~ /^reporter\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		$error = 0;		
		chomp($seq[$1] = <SEQ_FILE>);	
	}

	# if this line defines a seesaw gate, assign recognition sequences associated this node and nodes connected to both sides
	if ($line =~ /^seesaw\s*\[\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {

		$error = 0;	

		if (!defined($seq[$1])) {	chomp($seq[$1] = <SEQ_FILE>);	}

		my @fanin = split /,/,$2;
		my @fanout = split /,/,$3;
		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {
				$fanin[$i] = $1;
			} else {$error = 1;}
			if (!defined($seq[$fanin[$i]])) {	chomp($seq[$fanin[$i]] = <SEQ_FILE>);	}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
			if (!defined($seq[$fanout[$i]])) {	chomp($seq[$fanout[$i]] = <SEQ_FILE>);	}
		}
	}

	# if this line defines an initial concentration, check the syntax
	if ($line =~ /^conc\s*\[\s*(w|g|th)/) {

		# a seesaw signal
		if ($line =~ /^conc\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			$error = 0;
		}

		# a seesaw gate with signal bound by the signal's left side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	$error = 0;	}
		}

		# a seesaw gate with signal bound by the signal's right side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	$error = 0;	}		
		}

		# a seesaw threshold with signal coming in from the right side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	$error = 0;	}
		}

		# a seesaw threshold with signal coming in from the left side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	$error = 0;	}		
		}
	}

	if ($error == 1) {
		print "Syntax error in line ",$nl,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$nl,":\n",$line;
	}
}

close(SEESAW_FILE);
close(SEQ_FILE);
close(ERROR_FILE);

open(SEESAW_FILE,"<$ARGV[0].seesaw.txt") || die "$!";

print SEESAW_SEQ_FILE "\n";

# write all assigned recognition sequences and their complements to the seesaw sequence file
for (my $i=0; $i<=$#seq; $i++) {
	if (defined($seq[$i])) {

		$seq_c[$i] = $seq[$i];
		$seq_c[$i] =~ s/A/X/g;
		$seq_c[$i] =~ s/T/A/g;
		$seq_c[$i] =~ s/X/T/g;
		$seq_c[$i] =~ s/G/X/g;
		$seq_c[$i] =~ s/C/G/g;
		$seq_c[$i] =~ s/X/C/g;
		$seq_c[$i] = reverse($seq_c[$i]);

		print SEESAW_SEQ_FILE "S",$i,"\t= ",$seq[$i],"\tS",$i,"*\t= ",$seq_c[$i],"\n";
	}
}
print SEESAW_SEQ_FILE "\n";

while (($error == 0) && ($line = <SEESAW_FILE>)) {

	# for each input, write the signal strand sequence to the seesaw sequence file
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/) {	

		my $i = $2;		# this input wire connects node i to j
		my $j = $3;
		my $name = "INPUT(".$1.")";

		if ($line =~ /\#\s((\w|\^)+)/) {	$name = $1;	}
		print SEESAW_SEQ_FILE $name,": w",$i,",",$j,"\t= S",$j," T S",$i,"\t= ",$seq[$j]," ",$toehold," ",$seq[$i],"\n";
		${$wire[$i]}[$j] = 1;
	}

	# for each input fanout gate, write the threshold, gate and fuel sequences to the seesaw sequence file
	if ($line =~ /^inputfanout\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {	

		my @fanout = split /,/,$3;

		print SEESAW_SEQ_FILE "Th",$2,",",$1,":",$1,"-t\t= S",$1,"\t= ",$seq[$1],"\n";
		print SEESAW_SEQ_FILE "Th",$2,",",$1,":",$1,"-b\t= s",$2,"* T* S",$1,"*\t= ",substr($seq_c[$2],8,7)," ",$toehold_c," ",$seq_c[$1],"\n";
		for (my $i=0; $i<=$#fanout; $i++) {	
			print SEESAW_SEQ_FILE "w",$1,",",$fanout[$i],"\t= S",$fanout[$i]," T S",$1,"\t= ",$seq[$fanout[$i]]," ",$toehold," ",$seq[$1],"\n";	
		}
		print SEESAW_SEQ_FILE "G",$1,"-b\t= T* S",$1,"* T*\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$1]," ",$toehold_c," ",$clamp_c,"\n";
		print SEESAW_SEQ_FILE "w",$1,",f","\t= Sf T S",$1,"\t= ",$Sf," ",$toehold," ",$seq[$1],"\n";

	}

	# for each seesaw AND/OR gate, write the threshold, gate and fuel sequences to the seesaw sequence file
	if ($line =~ /^seesaw(AND|OR)\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/){	

		my @fanout = split /,/,$5;

		print SEESAW_SEQ_FILE "w",$2,",",$3,"\t= S",$3," T S",$2,"\t= ",$seq[$3]," ",$toehold," ",$seq[$2],"\n";
		print SEESAW_SEQ_FILE "G",$2,"-b\t= T* S",$2,"* T*\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$2]," ",$toehold_c," ",$clamp_c,"\n";
		print SEESAW_SEQ_FILE "Th",$2,",",$3,":",$3,"-t\t= S",$3,"\t= ",$seq[$3],"\n";
		print SEESAW_SEQ_FILE "Th",$2,",",$3,":",$3,"-b\t= s",$2,"* T* S",$3,"*\t= ",substr($seq_c[$2],8,7)," ",$toehold_c," ",$seq_c[$3],"\n";
		for (my $i=0; $i<=$#fanout; $i++) {	
			print SEESAW_SEQ_FILE "w",$3,",",$fanout[$i],"\t= S",$fanout[$i]," T S",$3,"\t= ",$seq[$fanout[$i]]," ",$toehold," ",$seq[$3],"\n";	
		}
		print SEESAW_SEQ_FILE "G",$3,"-b\t= T* S",$3,"* T*\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$3]," ",$toehold_c," ",$clamp_c,"\n";
		print SEESAW_SEQ_FILE "w",$3,",f","\t= Sf T S",$3,"\t= ",$Sf," ",$toehold," ",$seq[$3],"\n";
	}

	# for each reporter, write reporter sequences to the seesaw sequence file
	if ($line =~ /^reporter\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		print SEESAW_SEQ_FILE "Rep",$1,"-t\t= ",$quenc[$n]," S",$1,"\t= /",$quencIDT[$n],"/ ",$seq[$1],"\n";
		print SEESAW_SEQ_FILE "Rep",$1,"-b\t= T* S",$1,"* ",$fluor[$n],"\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$1]," /",$fluorIDT[$n],"/\n";
		$n++;
	}

	# for each defined initial concentration, write the corresponding sequence to the seesaw sequence file
	if ($line =~ /^conc\[(w|g|th)/) {
		
		# a seesaw signal
		if ($line =~ /^conc\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if (!defined(${$wire[$1]}[$2])) {
				print SEESAW_SEQ_FILE "w",$1,",",$2,"\t= S",$2," T S",$1,"\t= ",$seq[$2]," ",$toehold," ",$seq[$1],"\n";
				${$wire[$1]}[$2] = 1;
			}
		}

		# a seesaw gate with signal bound by the signal's left side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				if (!defined(${$wire[$2]}[$3])) {
					print SEESAW_SEQ_FILE "w",$2,",",$3,"\t= S",$3," T S",$2,"\t= ",$seq[$3]," ",$toehold," ",$seq[$2],"\n";
					${$wire[$2]}[$3] = 1;
				}
				if (!defined($gate[$2])) {
					print SEESAW_SEQ_FILE "G",$2,"-b\t= T* S",$2,"* T*\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$2]," ",$toehold_c," ",$clamp_c,"\n";
					$gate[$2] = 1;
				}
			}
		}

		# a seesaw gate with signal bound by the signal's right side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {		
				if (!defined(${$wire[$1]}[$2])) {
					print SEESAW_SEQ_FILE "w",$1,",",$2,"\t= S",$2," T S",$1,"\t= ",$seq[$2]," ",$toehold," ",$seq[$1],"\n";
					${$wire[$1]}[$2] = 1;
				}
				if (!defined($gate[$2])) {
					print SEESAW_SEQ_FILE "G",$2,"-b\t= T* S",$2,"* T*\t= ",$clamp_c," ",$toehold_c," ",$seq_c[$2]," ",$toehold_c," ",$clamp_c,"\n";
					$gate[$2] = 1;
				}
			}
		}

		# a seesaw threshold with signal coming in from the right side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {		
				print SEESAW_SEQ_FILE "Th",$2,":",$2,",",$3,"-t\t= S",$2,"\t= ",$seq[$2],"\n";
				print SEESAW_SEQ_FILE "Th",$2,":",$2,",",$3,"-b\t= S",$2,"* T* s",$3,"*\t= ",$seq_c[$2]," ",$toehold_c," ",substr($seq_c[$3],0,7),"\n";
			}
		}


		# a seesaw threshold with signal coming in from the left side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				print SEESAW_SEQ_FILE "Th",$1,",",$2,":",$2,"-t\t= S",$2,"\t= ",$seq[$2],"\n";
				print SEESAW_SEQ_FILE "Th",$1,",",$2,":",$2,"-b\t= s",$1,"* T* S",$2,"*\t= ",substr($seq_c[$1],8,7)," ",$toehold_c," ",$seq_c[$2],"\n";
			}
		}
	}
}

close(SEESAW_FILE);
close(SEESAW_SEQ_FILE);