#!/usr/bin/perl

use strict;
use warnings;
use File::Copy;

my $line;			# each line read from the input circuit file
my $n = 0;			# the number of this line
my $error = 0;		# if there is a syntax error in this line
my $def = 1;		# if copy the definitions from the example DSD file
my @seesaw;			# list of seesaw nodes	
my @inputfanout;	# list of input fanout gates
my @reporter;		# list of reporters
my @reporterREST;	# list of reporters with signal restoration
my @input;			# list of input signals
my @output;			# list of output signals
my @conc;			# list of initial concentrations
my %subroutine;		# list of subroutines in the DSD file

# open a seesaw circuit file for read
open(SEESAW_FILE,"<$ARGV[0].seesaw.txt") || die "$!";	

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

# open the seesaw DSD example file for read
open(EX_FILE,"<example.dna") || die "$!";

# open a seesaw DSD file for write
open(DSD_FILE,">$ARGV[0].seesaw.dna") || die "$!";

# read the seesaw circuit file
while (($error == 0) && ($line = <SEESAW_FILE>)) {

	$n++;
	$error = 1;

	# if this line is comment, skip it
	if ($line =~ /^(\#)/) {
		$error = 0;
	}

	# if this line is empty, skip it
	if ($line =~ /^(\s+)$/) {
		$error = 0;
	}

	# if this line defines a seesaw gate, check syntax
	if ($line =~ /^seesaw\s*\[\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {

		$error = 0;

		my @fanin = split /,/,$2;
		my @fanout = split /,/,$3;

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] !~ /^\s*(\d+)\s*$/) {	$error = 1;	}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] !~ /^\s*(\d+)\s*$/) {	$error = 1;	}
		}
	}

	# if this line defines an input
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/) {	
		$error = 0;

		my $i = $2;		# this input wire connects seesaw gate i to j
		my $j = $3;
		my $signal = "S".$i."L,S".$i.",S".$i."R,S".$j."L,S".$j.",S".$j."R";

		if ($line =~ /\#\s*(\w+)\s*\^\s*(0|1)/) {	# a dualrail input with notes
			if ($2 == 0) {	push(@input,"signal(ON,".$signal.") (* ".$1."^".$2." *)");	}
			if ($2 == 1) {	push(@input,"signal(OFF,".$signal.") (* ".$1."^".$2." *)");	}				
		} elsif ($line =~ /\#\s*(\w+)\s*/) {	# a regular input with notes
			push(@input,"signal(OFF,".$signal.") (* ".$1." *)");
		} else {							# a regular input without notes
			push(@input,"signal(OFF,".$signal.") (* input w[".$i.",".$j."] *)");
		}
	}

	# if this line defines an output
	if (($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*Fluor\s*\[\s*(\d+)\s*\](\s*|\s*\#.*)$/) ||
		($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/)) {	
		$error = 0;

		my $signal;
		my $note;

		# an output signal from a fluorescent reporter
		if ($line =~ /Fluor\s*\[\s*(\d+)\s*\]/) {	
			$signal = "<_ _ _ Fluor".$1.">";	
			$note = "output: Fluor[".$1."]";
		}

		# an output seesaw signal
		if ($line =~ /w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]/) {	
			$signal = "<S".$1."L^ S".$1." S".$1."R^ t^ S".$2."L^ S".$2." S".$2."R^>";	
			$note = "output w[".$1.",".$2."]";
		}
		
		# copy original notes or add new notes
		if ($line =~ /\#\s*((\w|\^)+)/) {		
			push(@output,$signal." (* ".$1." *)");
		} else {
			push(@output,$signal." (* ".$note." *)");
		}
	}

	# if this line defines the initial concentration of a seesaw signal, gate or threshold
	if ($line =~ /^conc\s*\[\s*(w|g|th)/) {
		
		# a seesaw signal
		if ($line =~ /^conc\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			$error = 0;	 
			push(@conc,"signal(".($3*10)."*N,S".$1."L,S".$1.",S".$1."R,S".$2."L,S".$2.",S".$2."R)");	
		}

		# a seesaw gate with signal bound by the signal's left side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;		
				push(@conc,"gateL(".($4*10)."*N,S".$2."L,S".$2.",S".$2."R,S".$3."L,S".$3.",S".$3."R)");	
			}
		}

		# a seesaw gate with signal bound by the signal's right side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;		
				push(@conc,"gateR(".($4*10)."*N,S".$1."L,S".$1.",S".$1."R,S".$2."L,S".$2.",S".$2."R)");	
			}
		}

		# a seesaw threshold with signal coming in from the right side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;	
				push(@conc,"thresholdR(".($4*10)."*N,S".$2."L,S".$2.",S".$2."R,S".$3."L)");	
			}
		}

		# a seesaw threshold with signal coming in from the left side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;
				push(@conc,"thresholdL(".($4*10)."*N,S".$1."R,S".$2."L,S".$2.",S".$2."R)");	
			}
		}
	}

	# if this line defines a seesaw AND/OR gate
	if ($line =~ /^seesaw(AND|OR)\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/){	
		$error = 0;

		my @fanin = split /,/,$4;
		my @fanout = split /,/,$5;
		my $subroutine = "seesaw".$1.($#fanin+1)."I".($#fanout+1)."O";
		
		$subroutine = $subroutine."(S".$2."L,S".$2.",S".$2."R,S".$3."L,S".$3.",S".$3."R";

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {	
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}

		$subroutine = $subroutine.",S".$fanout[0]."L,S".$fanout[0].",S".$fanout[0]."R";
		for (my $i=1; $i<=$#fanout; $i++) {	
			$subroutine = $subroutine.",S".$fanout[$i]."L,S".$fanout[$i].",S".$fanout[$i]."R";
		}
		$subroutine = $subroutine.")";

		push(@seesaw,$subroutine);	
	}

	# if this line defines a input fanout gate
	if ($line =~ /^inputfanout\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {	
		$error = 0;	

		my @fanout = split /,/,$3;
		my $subroutine = "inputfanout".($#fanout+1)."O";

		$subroutine = $subroutine."(S".$2."R,S".$1."L,S".$1.",S".$1."R";

		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}

		$subroutine = $subroutine.",S".$fanout[0]."L,S".$fanout[0].",S".$fanout[0]."R";
		for (my $i=1; $i<=$#fanout; $i++) {	
			$subroutine = $subroutine.",S".$fanout[$i]."L,S".$fanout[$i].",S".$fanout[$i]."R";
		}
		$subroutine = $subroutine.")";
		
		push(@inputfanout,$subroutine);	
	}

	# if this line defines a reporter gate
	if ($line =~ /^reporter\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		$error = 0;	
		push(@reporter,"reporter(15*N,S".$1."L,S".$1.",S".$1."R,Fluor".$1.")");	
	}

	# if this line defines a seesaw LTU gate
	if ($line =~ /^seesawLTU\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*,\s*(\d+|\d+\.\d+)\s*,\s*(\d+|\d+\.\d+)\s*,\s*\{([\s|\d|.|,]+)\}\s*\]\s*$/) {	
		$error = 0;

		my @fanin = split /,/,$3;
		my @fanout = split /,/,$4;
		my $sum = $5;
		my $threshold = $6;
		my @weight = split /,/,$7;
		my $subroutine = "seesawLTU".($#fanin+1)."I".($#fanout+1)."O";
		
		$subroutine = $subroutine."(S".$1."L,S".$1.",S".$1."R,S".$2."L,S".$2.",S".$2."R";

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {	
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#weight; $i++) {	
			if ($weight[$i] =~ /^\s*(\d+|\d+\.\d+)\s*$/) {
				$weight[$i] = $1;
			} else {$error = 1;}
		}
		
		for (my $i=0; $i<=$#fanout; $i++) {	
			$subroutine = $subroutine.",S".$fanout[$i]."L,S".$fanout[$i].",S".$fanout[$i]."R";
		}
		$subroutine = $subroutine.",".(10*$sum).",".(10*$threshold);
		for (my $i=0; $i<=$#weight; $i++) {	
			$subroutine = $subroutine.",".(10*$weight[$i]);
		}
		$subroutine = $subroutine.")";

		push(@seesaw,$subroutine);	
	}

	# if this line defines a reporter with signal restoration
	if ($line =~ /^reporterREST\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		$error = 0;	
		push(@reporterREST,"reporterREST(S".$1."L,S".$1.",S".$1."R,S".$2."L,S".$2.",S".$2."R,S".$3."L,S".$3.",S".$3."R)");
		push(@reporter,"reporter(15*N,S".$3."L,S".$3.",S".$3."R,Fluor".$3.")");	
	}

	if ($error == 1) {
		print "Syntax error in line ",$n,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$n,":\n",$line;
	}
}

close(SEESAW_FILE);
close(ERROR_FILE);

$line = <EX_FILE>;	# read the seesaw DSD example file

# copy the definitions to the new seesaw DSD file
while ($def == 1) {

	if ($line =~ /^directive plot/) {	# plot the outputs
		for (my $i=0; $i<=$#output; $i++) { print DSD_FILE "directive plot ",$output[$i],"\n";	}
	} else {
		print DSD_FILE $line;
	}

	$line = <EX_FILE>;
	if ($line =~ /^\(\* two-input/) {	$def = 0;	}	# end of definitions
}

close(EX_FILE);

# write seesaw AND/OR/LTU gate subroutines
for (my $i=0; $i<=$#seesaw; $i++) {	
	if (($seesaw[$i] =~ /^seesaw(AND|OR|LTU)(\d+)I(\d+)O/) && !defined(${${$subroutine{$1}}{$2}}{$3})) {
		print DSD_FILE "(* ",$2,"-input ",$3,"-output seesaw ",$1," gate *)\n";
		print DSD_FILE "def seesaw",$1,$2,"I",$3,"O(i1L,i1,i1R,i2L,i2,i2R";
		for (my $j=1; $j<=$3; $j++) {	print DSD_FILE ",k",$j,"L,k",$j,",k",$j,"R";	}
		if ($1 eq "LTU") {
			print DSD_FILE ",ws,th";
			for (my $j=1; $j<=$3; $j++) {	print DSD_FILE ",w",$j;		}
		}
		print DSD_FILE ") =\n( gateL(";
		if (($1 eq "AND") || ($1 eq "OR")) {
			print DSD_FILE $2*10;
		} else {
			print DSD_FILE "ws";
		}
		print DSD_FILE "*N,i1L,i1,i1R,i2L,i2,i2R)\n| thresholdL(";
		if ($1 eq "AND") {	print DSD_FILE ($2-1+0.2)*10;	} 
		elsif ($1 eq "OR") {		print DSD_FILE 6;	} 
		else{	print DSD_FILE "th";	}
		print DSD_FILE "*N,i1R,i2L,i2,i2R)\n";
		if (($1 eq "AND") || ($1 eq "OR")) {
			for (my $j=1; $j<=$3; $j++) {	print DSD_FILE "| gateL(10*N,i2L,i2,i2R,k",$j,"L,k",$j,",k",$j,"R)\n";	}
			print DSD_FILE "| signal(",$3*20,"*N,i2L,i2,i2R,fL,f,fR))\n\n";
		} else {
			for (my $j=1; $j<=$3; $j++) {	print DSD_FILE "| gateL(w",$j,"*N,i2L,i2,i2R,k",$j,"L,k",$j,",k",$j,"R)\n";	}
			print DSD_FILE "| signal((w1";
			for (my $j=2; $j<=$3; $j++) {	print DSD_FILE "+w",$j;	}
			print DSD_FILE ")*2*N,i2L,i2,i2R,fL,f,fR))\n\n";
		}
		${${$subroutine{$1}}{$2}}{$3} = 1;
	}
}

# write input fanout gate subroutines
for (my $i=0; $i<=$#inputfanout; $i++) {	
	if (($inputfanout[$i] =~ /^inputfanout(\d+)O/) && !defined(${$subroutine{"inputfanout"}}{$1})) {
		print DSD_FILE "(* input fanout gate with ",$1," outputs *)\n";
		print DSD_FILE "def inputfanout",$1,"O(jR,iL,i,iR";
		for (my $j=1; $j<=$1; $j++) {	print DSD_FILE ",k",$j,"L,k",$j,",k",$j,"R";	}
		print DSD_FILE ") =\n( thresholdL(2*N,jR,iL,i,iR)\n";
		for (my $j=1; $j<=$1; $j++) {	print DSD_FILE "| gateL(10*N,iL,i,iR,k",$j,"L,k",$j,",k",$j,"R)\n";	}
		print DSD_FILE "| signal(",$1*20,"*N,iL,i,iR,fL,f,fR))\n\n";
		${$subroutine{"inputfanout"}}{$1} = 1;
	}
}

# write reporter with signal restoration subroutine
print DSD_FILE "(* reporter with signal restoration *)\n";
print DSD_FILE "def reporterREST(i1L,i1,i1R,i2L,i2,i2R,i3L,i3,i3R) =\n";
print DSD_FILE "( gateL(10*N,i1L,i1,i1R,i2L,i2,i2R)\n";
print DSD_FILE "| thresholdL(4*N,i1R,i2L,i2,i2R)\n";
print DSD_FILE "| gateL(10*N,i2L,i2,i2R,i3L,i3,i3R)\n";
print DSD_FILE "| signal(20*N,i2L,i2,i2R,fL,f,fR))\n\n";

# write input signals, gates, thresholds and reporters to the seesaw DSD file
print DSD_FILE "( ",$input[0],"\n";
for (my $i=1; $i<=$#input; $i++) {	print DSD_FILE "| ",$input[$i],"\n";	}
for (my $i=0; $i<=$#seesaw; $i++) {	print DSD_FILE "| ",$seesaw[$i],"\n";	}
for (my $i=0; $i<=$#inputfanout; $i++) {	print DSD_FILE "| ",$inputfanout[$i],"\n";	}
for (my $i=0; $i<=$#conc; $i++) {	print DSD_FILE "| ",$conc[$i],"\n";	}
for (my $i=0; $i<=$#reporterREST; $i++) {	print DSD_FILE "| ",$reporterREST[$i],"\n";	}
for (my $i=0; $i<=$#reporter; $i++) {	print DSD_FILE "| ",$reporter[$i],"\n";	}
print DSD_FILE ")\n";

close(DSD_FILE);