#!/usr/bin/perl

use strict;
use warnings;
use File::Copy;

my $line;			# each line read from the input circuit file
my $n = 0;			# the number of this line
my $error = 0;		# if there is a syntax error in this line
my $W = 0;			# total initial concentration of signal strands			
my $G = 0;			# total initial concentration of gate complexes
my $TH = 0;			# total initial concentration of threshold complexes
my @seesaw;			# list of seesaw nodes	
my @inputfanout;	# list of input fanout gates
my @reporter;		# list of reporter gates
my @input;			# list of input signals
my @output;			# list of output signals
my @conc;			# list of initial concentrations
my @inputname;		# list of names of input signals
my @outputname;		# list of names of ourput signals

# open a seesaw circuit file for read
open(SEESAW_FILE,"<$ARGV[0].seesaw.txt") || die "$!";	

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

# open the CRN simulator Mathematica file for read
open(CRN_SIM_FILE,"<crn_simulator.m") || die "$!";

# open the seesaw simulator Mathematica file for read
open(SEESAW_SIM_FILE,"<seesaw_simulator.m") || die "$!";

# open the example Mathematica file for read
open(EX_FILE,"<example.m") || die "$!";

# open a seesaw Mathematica file for write
open(ODE_FILE,">$ARGV[0].seesaw.m") || die "$!";

while (($error == 0) && ($line = <SEESAW_FILE>)) {

	$n++;
	$error = 1;

	# if this line is comment, skip it
	if ($line =~ /^(\#)/) {
		$error = 0;
	}

	# if this line is empty, skip it
	if ($line =~ /^(\s+)$/) {
		$error = 0;
	}

	# if this line defines an input
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/) {	
		$error = 0;

		my $i = $2;		# this input wire connects node i to j
		my $j = $3;

		if ($line =~ /\#/) {						# if this input has a noted name
			if ($line =~ /\#\s*(\w+)\s*\^\s*(0|1)/) {		# if this is a dualrail input 
				if ($2 == 0) {	push(@input,"conc[w[".$i.",".$j."],(1-".$1.")*c]");	}
				else {	push(@input,"conc[w[".$i.",".$j."],".$1."*c]");	push(@inputname,$1); }				
			} elsif ($line =~ /\#\s*(\w+)\s*/) {		# if this is a regular input
				push(@input,"conc[w[".$i.",".$j."],".$1."*c]");
				push(@inputname,$1);
			}
		} else {		# if this input doesn't have a noted name, name it as x1, x2, etc.
			push(@input,"conc[w[".$i.",".$j."],x".($#input+2)."*c]");
			push(@inputname,"x".($#input+1));
		}
	}

	# if this line defines an output
	if (($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*Fluor\s*\[\s*(\d+)\s*\](\s*|\s*\#.*)$/) ||
		($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/)) {	
		$error = 0;
		
		# an output signal from a fluorescent reporter
		if ($line =~ /Fluor\s*\[\s*(\d+)\s*\]/) {	push(@output,"Fluor[".$1."]");	}
		
		# an output seesaw signal
		if ($line =~ /w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]/) {	push(@output,"w[".$1.",".$2."]");	}

		# save the noted name or name it as the signal 
		if ($line =~ /\#\s*((\w|\^)+)/) {	push(@outputname,$1)	}
		else {	push(@outputname,$output[-1])	}
	}

	# if this line defines a seesaw gate
	if ($line =~ /^seesaw\s*\[\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {

		$error = 0;

		my $seesawnode = "seesaw[".$1.",{";
		my @fanin = split /,/,$2;
		my @fanout = split /,/,$3;

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}
		
		$seesawnode = $seesawnode.$fanin[0];
		for (my $i=1; $i<=$#fanin; $i++) {	$seesawnode = $seesawnode.",".$fanin[$i];	}
		$seesawnode = $seesawnode."},{".$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$seesawnode = $seesawnode.",".$fanout[$i];	}
		$seesawnode = $seesawnode."}]";

		push(@seesaw,$seesawnode);
	}

	# if this line defines the initial concentration of a seesaw signal, gate or threshold
	if ($line =~ /^conc\s*\[\s*(w|g|th)/) {
		
		# a seesaw signal
		if ($line =~ /^conc\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			$error = 0;	 
			$W += $3;
			push(@conc,"conc[w[".$1.",".$2."],".$3."*c]");	
		}

		# a seesaw gate with signal bound by the signal's left side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;
				$G += $4;
				push(@conc,"conc[g[".$1.",w[".$2.",".$3."]],".$4."*c]");
			}
		}

		# a seesaw gate with signal bound by the signal's right side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;		
				$G += $4;
				push(@conc,"conc[g[w[".$1.",".$2."],".$3."],".$4."*c]");
			}
		}

		# a seesaw threshold with signal coming in from the right side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;	
				$TH += $4;
				push(@conc,"conc[th[".$1.",w[".$2.",".$3."]],".$4."*c]");
			}
		}

		# a seesaw threshold with signal coming in from the left side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;
				$TH += $4;
				push(@conc,"conc[th[w[".$1.",".$2."],".$3."],".$4."*c]");
			}
		}
	}

	# if this line defines a seesaw AND/OR gate
	if ($line =~ /^seesaw(AND|OR)\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/){	
		$error = 0;

		my $seesawnode = "seesaw".$1."[".$2.",".$3.",{";
		my @fanin = split /,/,$4;
		my @fanout = split /,/,$5;

		$W += 2*($#fanout + 1);
		$G += ($#fanin + $#fanout + 2);
		if ($1 eq "AND") {	$TH += ($#fanin + 0.2);	}	else {	$TH += 0.6;	}
		
		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}
		
		$seesawnode = $seesawnode.$fanin[0];
		for (my $i=1; $i<=$#fanin; $i++) {	$seesawnode = $seesawnode.",".$fanin[$i];	}
		$seesawnode = $seesawnode."},{".$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$seesawnode = $seesawnode.",".$fanout[$i];	}
		$seesawnode = $seesawnode."}]";

		push(@seesaw,$seesawnode);
	}

	# if this line defines a input fanout gate
	if ($line =~ /^inputfanout\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {	
		$error = 0;	

		my $inputfanoutnode = "inputfanout[".$1.",".$2.",{";
		my @fanout = split /,/,$3;

		$W += 2*($#fanout + 1);
		$G += ($#fanout + 1);
		$TH += 0.2;

		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}

		$inputfanoutnode = $inputfanoutnode.$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$inputfanoutnode = $inputfanoutnode.",".$fanout[$i];	}
		$inputfanoutnode = $inputfanoutnode."}]";

		push(@inputfanout,$inputfanoutnode);	
	}

	# if this line defines a reporter gate
	if ($line =~ /^reporter\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		$error = 0;			
		$G += 1.5;
		push(@reporter,"reporter[".$1.",".$2."]");	
	}

	if ($error == 1) {
		print "Syntax error in line ",$n,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$n,":\n",$line;
	}
}

close(SEESAW_FILE);
close(ERROR_FILE);

# copy the CRN simulator to the seesaw Mathematica file
while ($line = <CRN_SIM_FILE>) {	print ODE_FILE $line;	}

# copy the seesaw simulator to the seesaw Mathematica file
while ($line = <SEESAW_SIM_FILE>) {	print ODE_FILE $line;	}

# read the example Mathematica file
while ($line = <EX_FILE>) {

	# write the input values to the seesaw Mathematica file
	if ($line =~ /^input\s=\s/) {
		print ODE_FILE "input = {{OFF}";
		for (my $i=1; $i<=$#inputname; $i++) { print ODE_FILE ",{OFF}";	}
		print ODE_FILE "};\n";
	}

	# write the circuit name to the seesaw Mathematica file
	elsif ($line =~ /^SIM(\w+)=Table/) {
		if ($ARGV[0] =~ /^(\w+)/) {
			print ODE_FILE "SIM",$1,"=Table[\n";
		}
	}
	
	# write the seesaw system to the seesaw Mathematica file
	elsif ($line =~ /^gatesys/) {
		print ODE_FILE $line;
		for (my $i=0; $i<=$#seesaw; $i++) {	print ODE_FILE $seesaw[$i],",\n";	}
		for (my $i=0; $i<=$#inputfanout; $i++) {	print ODE_FILE $inputfanout[$i],",\n";	}
		for (my $i=0; $i<=$#reporter; $i++) {	print ODE_FILE $reporter[$i],",\n";	}
		for (my $i=0; $i<=$#conc; $i++) {	print ODE_FILE $conc[$i],",\n";	}
		for (my $i=0; $i<=$#input; $i++) {	print ODE_FILE $input[$i],",\n";	}
		if ($#inputname == ($#input - 1)/2) {	
			print ODE_FILE "conc[W,",$W+$#inputname+1,"*c],\n";	
		} else {	
			print ODE_FILE "conc[W,(",$W;
			for (my $i=0; $i<=$#inputname; $i++) {	print ODE_FILE "+",$inputname[$i];	}
			print ODE_FILE ")*c],\n";
		}
		print ODE_FILE "conc[G,",$G,"*c],\n";
		print ODE_FILE "conc[TH,1.1*",$TH,"*c]\n";
	}

	# skip the seesaw system from the example Mathematica file
	elsif ($line =~ /^(seesaw|inputfanout|reporter|conc)/) {}

	# write the outputs to the seesaw Mathematica file
	elsif ($line =~ /\/c\/\.sol/) {
		print ODE_FILE "{",$output[0],"[t]/c/.sol";
		for (my $i=1; $i<=$#output; $i++) {	print ODE_FILE ",",$output[$i],"[t]/c/.sol";	}
		print ODE_FILE "},\n";
	}

	# write the input names to the seesaw Mathematica file
	elsif ($line =~ /input\[\[/) {
		print ODE_FILE "{",$inputname[0],",input[[1]]}";
		for (my $i=1; $i<=$#inputname; $i++) {	print ODE_FILE ",{",$inputname[$i],",input[[",$i+1,"]]}"	}
		print ODE_FILE "\n";
	}

	# write the plot command to the seesaw Mathematica file
	elsif ($line =~ /^Plot\[/) {
		if ($ARGV[0] =~ /^(\w+)/) {
			print ODE_FILE "Needs[\"PlotLegends\`\"]\n";
			print ODE_FILE "Plot[SIM",$1,"//Flatten,{t,0,time}";
			print ODE_FILE ",Frame->True,FrameLabel->";
			print ODE_FILE "{Style[\"Time (hours)\",FontSize->30,FontFamily->\"Helvetica\"]";
			print ODE_FILE ",Style[\"Output\",FontSize->30,FontFamily->\"Helvetica\"]}";
			print ODE_FILE ",FrameTicks->{{Automatic,Automatic},{{{0,0},{7200,2},{14400,4},{21600,6},{28800,8},{36000,10}},None}}";
			print ODE_FILE ",PlotStyle->{Thick}";
			print ODE_FILE ",PlotLegend->{Style[\"",$outputname[0],"\",FontSize->20,FontFamily->\"Helvetica\"]";
			for (my $i=1; $i<=$#outputname; $i++) {	
				print ODE_FILE ",Style[\"",$outputname[$i],"\",FontSize->20,FontFamily->\"Helvetica\"]";	
			}
			print ODE_FILE "},LegendShadow->None,LegendSize->0.5,LegendPosition->{0.45,-0.2}";
			print ODE_FILE ",LabelStyle->Directive[FontSize->24,FontFamily->\"Helvetica\"]";
			print ODE_FILE ",PlotRange->All,AspectRatio->1/1.32,ImageSize->500]\n";
		}
	}

	# copy the rest from the example Mathematica file to the seesaw Mathematica file
	else {print ODE_FILE $line;}
}

close(CRN_SIM_FILE);
close(SEESAW_SIM_FILE);
close(EX_FILE);
close(ODE_FILE);