#!/usr/bin/perl

use strict;
use warnings;

my $sidereactions = "ON";			# side reacions are included ("ON") or not("OFF")
									# setting to "OFF" will greatly speed up simulation

my $line;			# each line read from the input circuit file
my $n = 0;			# the number of this line
my $error = 0;		# if there is a syntax error in this line

my $W = 0;							# total initial concentration of signal strands			
my $G = 0;							# total initial concentration of gate complexes
my $TH = 0;							# total initial concentration of threshold complexes
my $c = 50e-9;						# standard concentration 1x
my $kf = 2000000;					# fast strand displacement rate with extended toehold
my $ks = 50000;						# slow strand displacement rate with universal toehold
my $krf = 26;						# fast strand dissociation rate with universal toehold 
my $krs = 1.3;						# slow strand dissociation rate with extended toehold
my $kl = 10;						# strand displacement leak rate 
my $OFF = 0.1;						# logic OFF value
my $ON = 0.9;						# logic ON value

my @seesaw;							# seesaw abstractions
my %conc;							# species with their initial concentrations
my @rxn;							# reactions

# open a seesaw circuit file for read
open(SEESAW_FILE,"<$ARGV[0].seesaw.txt") || die "$!";

# open the syntax error file for write
open(ERROR_FILE,">error.txt") || die "$!";

while (($error == 0) && ($line = <SEESAW_FILE>)){

	$n++;
	$error = 1;

	# if this line is comment, skip it
	if ($line =~ /^(\#)/) {
		$error = 0;
	}

	# if this line is empty, skip it
	if ($line =~ /^(\s+)$/) {
		$error = 0;
	}

	# set initial concentrations of the inputs according to the logic OFF value
	if ($line =~ /^INPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/) {	
		$error = 0;

		my $i = $2;
		my $j = $3;

		if ($line =~ /\#\s*(\w+)\s*\^\s*0/) {	$conc{"w".$i."_".$j} = $ON;	}
		else {	$conc{"w".$i."_".$j} = $OFF;	}
	}

	# if this line defines an output, check syntex
	if (($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*Fluor\s*\[\s*(\d+)\s*\](\s*|\s*\#.*)$/) ||
		($line =~ /^OUTPUT\s*\(\s*(\d+)\s*\)\s*=\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\](\s*|\s*\#.*)$/)) {	
		$error = 0;
	}

	# translate each seesaw AND/OR gate into two seesaw nodes
	if ($line =~ /^seesaw(AND|OR)\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/){	
		$error = 0;

		my $type = $1;
		my $firstnode = $2;
		my $secondnode = $3;
		my @fanin = split(/,/,$4);
		my @fanout = split(/,/,$5);		

		$W += 2*($#fanout + 1);
		$G += ($#fanin + $#fanout + 2);
		if ($type eq "AND") {	$TH += 1.1 * ($#fanin + 0.2);	}	else {	$TH += 1.1 * 0.6;	}

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}

		my $seesawnode1 = "seesaw[".$firstnode.",{".$fanin[0];
		for (my $i=1; $i<=$#fanin; $i++) {	$seesawnode1 = $seesawnode1.",".$fanin[$i];	}
		$seesawnode1 = $seesawnode1."},{".$secondnode."}]";

		my $seesawnode2 = "seesaw[".$secondnode.",{".$firstnode."},{".$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$seesawnode2 = $seesawnode2.",".$fanout[$i];	}
		$seesawnode2 = $seesawnode2.",f}]";

		push(@seesaw,$seesawnode1);
		push(@seesaw,$seesawnode2);

		$conc{"G".$firstnode."__".$firstnode."_".$secondnode} = $#fanin + 1;
		for (my $i=0; $i<=$#fanout; $i++) {	$conc{"G".$secondnode."__".$secondnode."_".$fanout[$i]} = 1;	}
		$conc{"w".$secondnode."_f"} = 2 * ($#fanout + 1);
		if ($1 eq "OR") {	$conc{"Th".$firstnode."_".$secondnode."__".$secondnode} = 1.1 * 0.6;	} 
		else {	$conc{"Th".$firstnode."_".$secondnode."__".$secondnode} = 1.1 * ($#fanin + 0.2);	}
	}

	# translate each input fanout into one seesaw node with fanout
	if ($line =~ /^inputfanout\s*\[\s*(\d+)\s*,\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {	
		$error = 0;	

		my $fanoutnode = $1;
		my $inputnode = $2;
		my @fanout = split /,/,$3;

		$W += 2*($#fanout + 1);
		$G += ($#fanout + 1);
		$TH += 1.1 * 0.2;

		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}

		my $seesawnode = "seesaw[".$fanoutnode.",{".$inputnode."},{".$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$seesawnode = $seesawnode.",".$fanout[$i];	}
		$seesawnode = $seesawnode.",f}]";

		push(@seesaw,$seesawnode);

		for (my $i=0; $i<=$#fanout; $i++) {	$conc{"G".$fanoutnode."__".$fanoutnode."_".$fanout[$i]} = 1;	}
		$conc{"w".$fanoutnode."_f"} = 2 * ($#fanout + 1);
		$conc{"Th".$inputnode."_".$fanoutnode."__".$fanoutnode} = 1.1 * 0.2;	
		
	}

	# translate each reporter into its corresponding reactions
	if ($line =~ /^reporter\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*$/) {
		$error = 0;	

		$G += 1.5;

		# reporting reaction
		push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$2."_".$1);
		push(@{${$rxn[$#rxn]}{"reac"}},"Rep".$1);
		push(@{${$rxn[$#rxn]}{"prod"}},"Fluor".$1);
		push(@{${$rxn[$#rxn]}{"rate"}},"ks");

		$conc{"Rep".$1} = 1.5;
		$conc{"Fluor".$1} = 0;

		if ($sidereactions eq "ON") {

			# universal toehold binding reactions
			push(@{${$rxn[$#rxn+1]}{"reac"}},"Rep".$1);
			push(@{${$rxn[$#rxn]}{"reac"}},"W");
			push(@{${$rxn[$#rxn]}{"prod"}},"Rep".$1."_W");
			push(@{${$rxn[$#rxn]}{"rate"}},"kf","krf");
	
			push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$2."_".$1);
			push(@{${$rxn[$#rxn]}{"reac"}},"G");
			push(@{${$rxn[$#rxn]}{"prod"}},"w".$2."_".$1."_G");
			push(@{${$rxn[$#rxn]}{"rate"}},"kf","krf");
	
			push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$2."_".$1);
			push(@{${$rxn[$#rxn]}{"reac"}},"TH");
			push(@{${$rxn[$#rxn]}{"prod"}},"w".$2."_".$1."_TH");
			push(@{${$rxn[$#rxn]}{"rate"}},"kf","krs");

			$conc{"Rep".$1."_W"} = 0;
			$conc{"w".$2."_".$1."_G"} = 0;
			$conc{"w".$2."_".$1."_TH"} = 0;
		}
	}

	# save each defined seesaw node
	if ($line =~ /^seesaw\s*\[\s*(\d+)\s*,\s*\{([\s|\d|,]+)\}\s*,\s*\{([\s|\d|,]+)\}\s*\]\s*$/) {

		$error = 0;

		my $seesawnode = "seesaw[".$1.",{";
		my @fanin = split /,/,$2;
		my @fanout = split /,/,$3;

		for (my $i=0; $i<=$#fanin; $i++) {	
			if ($fanin[$i] =~ /^\s*(\d+)\s*$/) {
				$fanin[$i] = $1;
			} else {$error = 1;}
		}
		for (my $i=0; $i<=$#fanout; $i++) {	
			if ($fanout[$i] =~ /^\s*(\d+)\s*$/) {
				$fanout[$i] = $1;
			} else {$error = 1;}
		}
		
		$seesawnode = $seesawnode.$fanin[0];
		for (my $i=1; $i<=$#fanin; $i++) {	$seesawnode = $seesawnode.",".$fanin[$i];	}
		$seesawnode = $seesawnode."},{".$fanout[0];
		for (my $i=1; $i<=$#fanout; $i++) {	$seesawnode = $seesawnode.",".$fanout[$i];	}
		$seesawnode = $seesawnode."}]";

		push(@seesaw,$seesawnode);
	}

	# save each defined initial concentration
	if ($line =~ /^conc\s*\[\s*(w|g|th)/) {
		
		# a seesaw signal
		if ($line =~ /^conc\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			$error = 0;	 
			$W += $3;
			$conc{"w".$1."_".$2} = $3;	
		}

		# a seesaw gate with signal bound by the signal's left side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;
				$G += $4;
				$conc{"G".$1."__".$2."_".$3} = $4;
			}
		}

		# a seesaw gate with signal bound by the signal's right side
		if ($line =~ /^conc\s*\[\s*g\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;		
				$G += $4;
				$conc{"G".$1."_".$2."__".$3} = $4;
			}
		}

		# a seesaw threshold with signal coming in from the right side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*(\d+)\s*,\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/){
			if ($1 == $2) {	
				$error = 0;	
				$TH += $4;
				$conc{"Th".$1."__",$2."_".$3} = $4;
			}
		}

		# a seesaw threshold with signal coming in from the left side 
		if ($line =~ /^conc\s*\[\s*th\s*\[\s*w\s*\[\s*(\d+)\s*,\s*(\d+)\s*\]\s*,\s*(\d+)\s*\]\s*,\s*(\d+|\d+\.\d+)\s*\*\s*c\s*\]\s*$/) {
			if ($2 == $3) {	
				$error = 0;
				$TH += $4;
				$conc{"Th".$1."_",$2."__".$3} = $4;
			}
		}
	}

	if ($error == 1) {
		print "Syntax error in line ",$n,":\n",$line;
		print ERROR_FILE "Syntax error in line ",$n,":\n",$line;
	}
}

close(SEESAW_FILE);
close(ERROR_FILE);

# translate seesaw abstractions into reactions
for (my $i=0; $i<=$#seesaw; $i++) {

	if ($seesaw[$i] =~ /^seesaw\[(\d+),\{(.*)\},\{(.*)\}\]/) {	

		my @fanin = split(/,/,$2);
		my @fanout = split(/,/,$3);

		# set all the undifined initial concentrations to be zero
		for (my $i=0; $i<=$#fanin; $i++) {	
			if (!defined($conc{"w".$fanin[$i]."_".$1})) {	$conc{"w".$fanin[$i]."_".$1} = 0;	}
			if (!defined($conc{"G".$fanin[$i]."_".$1."__".$1})) {	$conc{"G".$fanin[$i]."_".$1."__".$1} = 0;	}
			if (!defined($conc{"Th".$fanin[$i]."_".$1."__".$1})) {	$conc{"Th".$fanin[$i]."_".$1."__".$1} = 0;	}
			if ($sidereactions eq "ON") {
				$conc{"w".$fanin[$i]."_".$1."_G"} = 0;
				$conc{"w".$fanin[$i]."_".$1."_TH"} = 0;
				$conc{"G".$fanin[$i]."_".$1."__".$1."_W"} = 0;
				$conc{"ThW_".$fanin[$i]."_".$1."__".$1} = 0;
			}
		}

		for (my $i=0; $i<=$#fanout; $i++) {	
			if (!defined($conc{"w".$1."_".$fanout[$i]})) {	$conc{"w".$1."_".$fanout[$i]} = 0;	}
			if (!defined($conc{"G".$1."__".$1."_".$fanout[$i]})) {	$conc{"G".$1."__".$1."_".$fanout[$i]} = 0;	}	
			if (!defined($conc{"Th".$1."__".$1."_".$fanout[$i]})) {	$conc{"Th".$1."__".$1."_".$fanout[$i]} = 0;	}
			if ($sidereactions eq "ON") {
				$conc{"w".$1."_".$fanout[$i]."_G"} = 0;
				$conc{"w".$1."_".$fanout[$i]."_TH"} = 0;
				$conc{"GW_".$1."__".$1."_".$fanout[$i]} = 0;
				$conc{"Th".$1."__".$1."_".$fanout[$i]."_W"} = 0;
			}
		}

		# toehold exchange reactions
		for (my $i=0; $i<=$#fanin; $i++) {
			for (my $j=0; $j<=$#fanout; $j++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$fanin[$i]."_".$1);
				push(@{${$rxn[$#rxn]}{"reac"}},"G".$1."__".$1."_".$fanout[$j]);
				push(@{${$rxn[$#rxn]}{"prod"}},"G".$fanin[$i]."_".$1."__".$1);
				push(@{${$rxn[$#rxn]}{"prod"}},"w".$1."_".$fanout[$j]);
				push(@{${$rxn[$#rxn]}{"rate"}},"ks","ks");
			}
		}

		# thresholding reactions
		for (my $i=0; $i<=$#fanin; $i++) {
			push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$fanin[$i]."_".$1);
			push(@{${$rxn[$#rxn]}{"reac"}},"Th".$fanin[$i]."_".$1."__".$1);
			push(@{${$rxn[$#rxn]}{"prod"}},"waste");
			push(@{${$rxn[$#rxn]}{"rate"}},"kf");
		}

		for (my $i=0; $i<=$#fanout; $i++) {
			push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$1."_".$fanout[$i]);
			push(@{${$rxn[$#rxn]}{"reac"}},"Th".$1."__".$1."_".$fanout[$i]);
			push(@{${$rxn[$#rxn]}{"prod"}},"waste");
			push(@{${$rxn[$#rxn]}{"rate"}},"kf");
		}

		if ($sidereactions eq "ON") {

			# universal toehold binding reactions
			for (my $i=0; $i<=$#fanin; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"G".$fanin[$i]."_".$1."__".$1);
				push(@{${$rxn[$#rxn]}{"reac"}},"W");
				push(@{${$rxn[$#rxn]}{"prod"}},"G".$fanin[$i]."_".$1."__".$1."_W");
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krf");
			}
	
			for (my $i=0; $i<=$#fanout; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"G".$1."__".$1."_".$fanout[$i]);
				push(@{${$rxn[$#rxn]}{"reac"}},"W");
				push(@{${$rxn[$#rxn]}{"prod"}},"GW_".$1."__".$1."_".$fanout[$i]);
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krf");
			}
	
			for (my $i=0; $i<=$#fanin; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"Th".$fanin[$i]."_".$1."__".$1);
				push(@{${$rxn[$#rxn]}{"reac"}},"W");
				push(@{${$rxn[$#rxn]}{"prod"}},"ThW_".$fanin[$i]."_".$1."__".$1);
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krs");
			}
	
			for (my $i=0; $i<=$#fanout; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"Th".$1."__".$1."_".$fanout[$i]);
				push(@{${$rxn[$#rxn]}{"reac"}},"W");
				push(@{${$rxn[$#rxn]}{"prod"}},"Th".$1."__".$1."_".$fanout[$i]."_W");
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krs");
			}
	
			for (my $i=0; $i<=$#fanin; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$fanin[$i]."_".$1);
				push(@{${$rxn[$#rxn]}{"reac"}},"G");
				push(@{${$rxn[$#rxn]}{"prod"}},"w".$fanin[$i]."_".$1."_G");
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krf");
			}
	
			for (my $i=0; $i<=$#fanin; $i++) {
				push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$fanin[$i]."_".$1);
				push(@{${$rxn[$#rxn]}{"reac"}},"TH");
				push(@{${$rxn[$#rxn]}{"prod"}},"w".$fanin[$i]."_".$1."_TH");
				push(@{${$rxn[$#rxn]}{"rate"}},"kf","krs");
			}
	
			# leak reactions
			for (my $i=0; $i<=$#fanin; $i++) {
				for (my $j=0; $j<=$#fanin; $j++) {
					push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$fanin[$i]."_".$1);
					push(@{${$rxn[$#rxn]}{"reac"}},"G".$fanin[$j]."_".$1."__".$1);
					push(@{${$rxn[$#rxn]}{"prod"}},"G".$fanin[$i]."_".$1."__".$1);
					push(@{${$rxn[$#rxn]}{"prod"}},"w".$fanin[$j]."_".$1);
					push(@{${$rxn[$#rxn]}{"rate"}},"kl");
				}
			}

			for (my $i=0; $i<=$#fanout; $i++) {
				for (my $j=0; $j<=$#fanout; $j++) {
					push(@{${$rxn[$#rxn+1]}{"reac"}},"w".$1."_".$fanout[$i]);
					push(@{${$rxn[$#rxn]}{"reac"}},"G".$1."__".$1."_".$fanout[$j]);
					push(@{${$rxn[$#rxn]}{"prod"}},"G".$1."__".$1."_".$fanout[$i]);
					push(@{${$rxn[$#rxn]}{"prod"}},"w".$1."_".$fanout[$j]);
					push(@{${$rxn[$#rxn]}{"rate"}},"kl");
				}
			}
		}
	}
}

$conc{"waste"} = 0;

if ($sidereactions eq "ON") {
	$conc{"W"} = $W;
	$conc{"G"} = $G;
	$conc{"TH"} = $TH;
}

# open a SBML file for write
open(SBML,">$ARGV[0].seesaw.sbml.xml") || die "$!";

# write the level, version, model name etc. to the SBML file
print SBML "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
print SBML "<sbml xmlns=\"http://www.sbml.org/sbml/level2\" level=\"2\" version=\"1\">\n";
print SBML "  <model name=\"$ARGV[0]\">\n";
print SBML "    <listOfCompartments>\n";
print SBML "      <compartment id=\"default\" name=\"default\" size=\"1\"/>\n";
print SBML "    </listOfCompartments>\n";

# write the species and their initial concentrations to the SBML file
print SBML "    <listOfSpecies>\n";
foreach my $name (sort keys %conc) {
	print SBML "      <species id=\"",$name,"\" name=\"",$name,"\" compartment=\"default\" initialAmount=\"",$conc{$name}*$c,"\"/>\n";
}
print SBML "    </listOfSpecies>\n";

# write the rate parameters to the SBML file
print SBML "    <listOfParameters>\n";
print SBML "      <parameter id=\"ks\" name=\"ks\" value=\"",$ks,"\"/>\n";
print SBML "      <parameter id=\"kf\" name=\"kf\" value=\"",$kf,"\"/>\n";
if ($sidereactions eq "ON") {
	print SBML "      <parameter id=\"krs\" name=\"krs\" value=\"",$krs,"\"/>\n";
	print SBML "      <parameter id=\"krf\" name=\"krf\" value=\"",$krf,"\"/>\n";
	print SBML "      <parameter id=\"kl\" name=\"kl\" value=\"",$kl,"\"/>\n";
}
print SBML "    </listOfParameters>\n";

# write the reactions to the SBML file
print SBML "    <listOfReactions>\n";
for(my $i=0; $i<=$#rxn; $i++){		
	print SBML "      <reaction id=\"R",$i+1,"\" name=\"R",$i+1,"\">\n";
	print SBML "        <listOfReactants>\n";
	for (my $j=0; $j<=$#{${$rxn[$i]}{"reac"}}; $j++) {
		print SBML "          <speciesReference species=\"",${${$rxn[$i]}{"reac"}}[$j],"\"/>\n";
	}
	print SBML "        </listOfReactants>\n";
	print SBML "        <listOfProducts>\n";
	for (my $j=0; $j<=$#{${$rxn[$i]}{"prod"}}; $j++) {
		print SBML "          <speciesReference species=\"",${${$rxn[$i]}{"prod"}}[$j],"\"/>\n";
	}
	print SBML "        </listOfProducts>\n";
	print SBML "        <kineticLaw>\n";
	print SBML "          <math xmlns=\"http://www.w3.org/1998/Math/MathML\">\n";
	print SBML "            <apply>\n";
	if ($#{${$rxn[$i]}{"rate"}} == 1) {
		print SBML "              <minus/>\n";
		print SBML "              <apply>\n";
		print SBML "                <times/>\n";
		print SBML "                <ci> ",${${$rxn[$i]}{"rate"}}[0]," </ci>\n";
		for (my $j=0; $j<=$#{${$rxn[$i]}{"reac"}}; $j++) {
			print SBML "                <ci> ",${${$rxn[$i]}{"reac"}}[$j]," </ci>\n";
		}
		print SBML "              </apply>\n";
		print SBML "              <apply>\n";
		print SBML "                <times/>\n";
		print SBML "                <ci> ",${${$rxn[$i]}{"rate"}}[1]," </ci>\n";
		for (my $j=0; $j<=$#{${$rxn[$i]}{"prod"}}; $j++) {
			print SBML "                <ci> ",${${$rxn[$i]}{"prod"}}[$j]," </ci>\n";
		}
		print SBML "              </apply>\n";
	} else {
		print SBML "              <times/>\n";
		print SBML "              <ci> ",${${$rxn[$i]}{"rate"}}[0]," </ci>\n";
		for (my $j=0; $j<=$#{${$rxn[$i]}{"reac"}}; $j++) {
			print SBML "              <ci> ",${${$rxn[$i]}{"reac"}}[$j]," </ci>\n";
		}
	}
	print SBML "            </apply>\n";
	print SBML "          </math>\n";
	print SBML "        </kineticLaw>\n";
	print SBML "      </reaction>\n";
}
print SBML "    </listOfReactions>\n";
print SBML "  </model>\n";
print SBML "</sbml>\n";

close(SBML);