(* ::Package:: *)

(* ::Section:: *)
(*Define CRN notation and data types*)


<<Notation`


Notation[\!\(\*
TagBox[
RowBox[{"r_", 
OverscriptBox["\[LongRightArrow]", "k_"], "p_"}],
"NotationTemplateTag"]\) \[DoubleLongLeftArrow] \!\(\*
TagBox[
RowBox[{"rxn", "[", 
RowBox[{"r_", ",", "p_", ",", "k_"}], "]"}],
"NotationTemplateTag"]\)]


rxn::about="Represents an irreversible reaction. eg. rxn[a+b,c,1]";
revrxn::about="Represents a reversible reaction. eg. revrxn[a+b,c,1,1]";
conc::about="Initial concentration: conc[{x,y},10]";


revrxn[r_,p_,k1_,k2_]:=Sequence[rxn[r,p,k1],rxn[p,r,k2]]


SetAttributes[{gate,rxn,revrxn}, HoldAll]


Seq:=Sequence (*to use instead of /.List->Sequence in definition of gates*)


conc[xs_List,c_]:=(conc[#,c]&/@xs)/.List->Sequence


(* ::Section:: *)
(*Define CRN functions*)


(* ::Text:: *)
(*Converts a system of gates with conc statements to NDSolve statement:*)


gatesysToNDSolve[gatesys_,endtime_]:=
Module[{
inputspecs=Cases[gatesys,conc[x_,c_]:>{x,c}],rsys=DeleteCases[gatesys,conc[___]]},
rxnsysToNDSolve[rsys,inputspecs,endtime]]


(* ::Text:: *)
(*Gets the set of all species for a reaction system:*)


spcsInRxnsys[rsys_]:=Cases[Cases[rsys,rxn[r_,p_,_]:>Seq[r,p]]/.Times|Plus->Seq,_Symbol|_Symbol[__]]//Union


(* ::Text:: *)
(*Converts a reaction system to a list of odes:*)


rxnsysToOdesys[rsys_]:=
Module[{spcs=spcsInRxnsys[rsys], rrates, spccoeffs,odes},
(*create terms for the rates of each reaction*)
rrates = rsys/.rxn[r_,_,k_]:>(k (r/.{Times[b_,s_]:>s^b,Plus->Times}));
(*for each species, get list of net coefficient for each reaction*)
spccoeffs=Coefficient[rsys/.rxn[r_,p_,_]:>p-r,#]& /@ spcs;
(*create ode for each species*)
odes=MapThread[Function[{spc,coeffs},HoldForm[D[spc,t]]==Total[coeffs*rrates]],{spcs, spccoeffs}];
(*change all species conc to be functions of t*)
odes/.s_/;MemberQ[spcs,s]:>s[t]]//ReleaseHold;


(* ::Text:: *)
(*Converts a reaction system to an NDSolve statement, and solves:*)


rxnsysToNDSolve[rsys_,inputspecs_, endtime_]:=
Module[{spcs=spcsInRxnsys[rsys],odesys=rxnsysToOdesys[rsys], NDSolveEqns},
NDSolveEqns = Join[odesys,(#[0]==Cases[inputspecs,{#,c0_}:>c0]/.{{c0_}:> c0,{}->0})&/@spcs];
Quiet[NDSolve[NDSolveEqns, spcs, {t,0,endtime},
AccuracyGoal->MachinePrecision,PrecisionGoal->MachinePrecision,WorkingPrecision->MachinePrecision,MaxSteps->Infinity],{NDSolve::"precw"}][[1]]]


(* ::Text:: *)
(*Generates list of species matching formal pattern:*)


spcsPattern[rsys_,pattern_]:=Cases[spcsInRxnsys[rsys],pattern]


(* ::Text:: *)
(*Generates list of species matching a pattern (eg "g$*" to match all implicit gates; can also do RegularExpression["o..d.\$.*"]):*)


spcsPatternString[rsys_,pattern_]:=Select[spcsInRxnsys[rsys],StringMatchQ[ToString[#],pattern]&]


(* ::Text:: *)
(*Generates set of initial values for given list of species:*)


initsSpcs[spcs_,value_]:={#,value}&/@spcs


(* ::Text:: *)
(*Generates set of initial values for species matching formal pattern:*)


initsSpcsPattern[rsys_,pattern_,value_]:={#,value}&/@spcsPattern[rsys,pattern]


(* ::Text:: *)
(*Generates set of initial values for species matching string pattern:*)


initsSpcsPatternString[rsys_,pattern_,value_]:={#,value}&/@spcsPatternString[rsys,pattern]
